"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ref_1 = require("../../../../utils/ref");
const Either_1 = require("fp-ts/lib/Either");
const serialized_type_1 = require("../../common/data/serialized-type");
const pipeable_1 = require("fp-ts/lib/pipeable");
const fp_ts_1 = require("fp-ts");
const either_utils_1 = require("@devexperts/utils/dist/adt/either.utils");
const operation_object_1 = require("./operation-object");
const types_1 = require("../../../../utils/types");
const serialized_dependency_1 = require("../../common/data/serialized-dependency");
const client_1 = require("../../common/bundled/client");
const schema_object_1 = require("./schema-object");
const Option_1 = require("fp-ts/lib/Option");
const option_utils_1 = require("@devexperts/utils/dist/adt/option.utils");
const function_1 = require("fp-ts/lib/function");
const string_1 = require("../../../../utils/string");
const serialized_parameter_1 = require("../../common/data/serialized-parameter");
const array_1 = require("../../../../utils/array");
exports.serializeChannelItemObject = (from, channel, channelItemObject, kind) => {
    const send = pipeable_1.pipe(channelItemObject.subscribe, fp_ts_1.option.map(operation => operation_object_1.serializeSubscribeOperationObject(from, channel, operation, kind)));
    const message = pipeable_1.pipe(channelItemObject.publish, fp_ts_1.option.map(operation => operation_object_1.serializePublishOperationObject(from, channel, operation)));
    const serialized = pipeable_1.pipe(fp_ts_1.array.compact([send, message]), either_utils_1.sequenceEither, fp_ts_1.either.map(serialized => serialized_type_1.intercalateSerializedTypes(serialized_type_1.serializedType(';', ',', [], []), serialized)));
    const ws = pipeable_1.pipe(channelItemObject.bindings, fp_ts_1.option.chain(bindings => bindings.ws));
    const method = pipeable_1.pipe(ws, fp_ts_1.option.chain(ws => ws.method), fp_ts_1.option.getOrElse(() => 'GET'));
    const query = pipeable_1.pipe(ws, fp_ts_1.option.chain(ws => ws.query), fp_ts_1.option.map(query => {
        const required = isRequired(query);
        return pipeable_1.pipe(validateSchemaObject(query), fp_ts_1.either.chain(query => schema_object_1.serializeSchemaObject(from, query)), fp_ts_1.either.map(serialized => serialized_type_1.getSerializedPropertyType('query', required, serialized)), fp_ts_1.either.map(serialized_parameter_1.fromSerializedType(required)));
    }));
    const headers = pipeable_1.pipe(ws, fp_ts_1.option.chain(ws => ws.headers), fp_ts_1.option.map(headers => {
        const required = isRequired(headers);
        return pipeable_1.pipe(validateSchemaObject(headers), fp_ts_1.either.chain(headers => schema_object_1.serializeSchemaObject(from, headers)), fp_ts_1.either.map(serialized => serialized_type_1.getSerializedPropertyType('headers', required, serialized)), fp_ts_1.either.map(serialized_parameter_1.fromSerializedType(required)));
    }));
    const parameters = pipeable_1.pipe(fp_ts_1.array.compact([query, headers]), either_utils_1.sequenceEither);
    return either_utils_1.combineEither(serialized, client_1.clientRef, parameters, (serialized, clientRef, parameters) => {
        const serializedParameters = serialized_parameter_1.intercalateSerializedParameters(serialized_parameter_1.serializedParameter(';', ',', false, [], []), parameters);
        const hasParameters = parameters.length > 0;
        const channelType = types_1.foldKind(kind, 'WebSocketClient', 'WebSocketClient1', 'WebSocketClient2');
        const type = `(${string_1.when(hasParameters, `parameters: { ${serializedParameters.type} }`)}) => { ${serialized.type} }`;
        const io = `(${string_1.when(hasParameters, 'parameters')}) => {
			${string_1.when(hasParameters, `const encoded = partial({ ${serializedParameters.io} }).encode(parameters);`)}
			const channel = e.webSocketClient.channel({
				channel: ${JSON.stringify(channel)},
				method: '${method}',
				${string_1.when(hasParameters, '...encoded,')}
			})
			return { ${serialized.io} };
		}`;
        const dependencies = array_1.concatIf(hasParameters, [
            ...serialized.dependencies,
            ...serializedParameters.dependencies,
            serialized_dependency_1.serializedDependency(channelType, ref_1.getRelativePath(from, clientRef)),
        ], [serialized_dependency_1.serializedDependency('partial', 'io-ts')]);
        const refs = [...serialized.refs, ...serializedParameters.refs];
        return pipeable_1.pipe(serialized_type_1.serializedType(type, io, dependencies, refs), serialized_type_1.getSerializedOptionPropertyType(`[${JSON.stringify(channel)}]`, true));
    });
};
const validateSchemaObject = (schemaObject) => Option_1.isSome(schemaObject.properties)
    ? Either_1.right(schemaObject)
    : Either_1.left(new Error(`ChannelItemObject.query and ChannelItemObject.header should have "properties" field`));
const isRequired = (objectSchemaObject) => pipeable_1.pipe(option_utils_1.sequenceTOption(objectSchemaObject.properties, objectSchemaObject.required), fp_ts_1.option.map(([properties, required]) => fp_ts_1.record.keys(properties).some(key => required.has(key))), fp_ts_1.option.getOrElse(function_1.constFalse));
