"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ref_1 = require("../../../../utils/ref");
const pipeable_1 = require("fp-ts/lib/pipeable");
const fp_ts_1 = require("fp-ts");
const fs_1 = require("../../../../utils/fs");
exports.utilsRef = ref_1.fromString('#/utils/utils');
const utils = `
	import { either } from 'fp-ts/lib/Either';
	import { Type, failure, success, string as tstring } from 'io-ts';

	export const DateFromISODateStringIO = new Type<Date, string, unknown>(
		'DateFromISODateString',
		(u): u is Date => u instanceof Date,
		(u, c) =>
			either.chain(tstring.validate(u, c), dateString => {
				const [year, calendarMonth, day] = dateString.split('-');
				const d = new Date(+year, +calendarMonth - 1, +day);
				return isNaN(d.getTime()) ? failure(u, c) : success(d);
			}),
		a =>
			\`\${a.getFullYear().toString().padStart(4, '0')}-\${(a.getMonth() + 1).toString().padStart(2, '0')}-\${a
				.getDate()
				.toString()
				.padStart(2, '0')}\`,
	);
`;
exports.utilsFile = pipeable_1.pipe(exports.utilsRef, fp_ts_1.either.map(ref => fs_1.fromRef(ref, '.ts', utils)));
